#pragma once
#include <score/widgets/SignalUtils.hpp>

#include <ossia/audio/jack_protocol.hpp>

#include <QFormLayout>
#include <QLabel>
#include <QSpinBox>
#include <QWidget>

#include <Audio/AudioInterface.hpp>
#include <Audio/Settings/Model.hpp>
#include <Audio/Settings/View.hpp>
#include <score/command/Dispatchers/SettingsCommandDispatcher.hpp>
#include <score/tools/Bind.hpp>

namespace Audio
{
#if defined(OSSIA_AUDIO_JACK)
class JackFactory final : public AudioFactory
{
  SCORE_CONCRETE("7ff2af00-f2f5-4930-beec-0e2d21eda195")
  private:
    std::weak_ptr<ossia::jack_client> m_client{};
public:
  ~JackFactory() override {}

  QString prettyName() const override { return QObject::tr("JACK"); }
  std::unique_ptr<ossia::audio_engine> make_engine(
      const Audio::Settings::Model& set,
      const score::ApplicationContext& ctx) override
  {
    static_assert(std::is_base_of_v<ossia::audio_engine, ossia::jack_engine>);
    auto clt = m_client.lock();
    if(!clt)
    {
      m_client = (clt = std::make_shared<ossia::jack_client>("ossia score"));
    }
    return std::make_unique<ossia::jack_engine>(
        clt,
        set.getDefaultIn(),
        set.getDefaultOut());
  }

  QWidget* make_settings(
      Audio::Settings::Model& m,
      Audio::Settings::View& v,
      score::SettingsCommandDispatcher& m_disp,
      QWidget* parent) override
  {
    auto w = new QWidget{parent};
    auto lay = new QFormLayout{w};

    using Model = Audio::Settings::Model;

    auto in_count = new QSpinBox{w};
    auto out_count = new QSpinBox{w};


#if defined(_WIN32)
    {
      if (!ossia::has_jackd_process())
      {
        qDebug() << "JACK server not running?";
        throw std::runtime_error("Audio error: no JACK server");
      }
    }
#endif

    qDebug() << "JACK: " << WeakJack::instance().available();
    std::shared_ptr<ossia::jack_client> client = m_client.lock();
    if(!client)
      m_client = (client = std::make_shared<ossia::jack_client>("ossia score"));

    {
      auto rate = jack_get_sample_rate(*client);
      lay->addRow(QObject::tr("Rate"), new QLabel{QString::number(rate)});
      m.setRate(rate);
    }
    {
      auto bs = jack_get_buffer_size(*client);
      lay->addRow(QObject::tr("Buffer size"), new QLabel{QString::number(bs)});
      m.setBufferSize(bs);
    }

    {
      in_count->setRange(0, 1024);
      lay->addRow(QObject::tr("Inputs"), in_count);
      QObject::connect(
          in_count,
          SignalUtils::QSpinBox_valueChanged_int(),
          w,
          [=, &m, &m_disp](int i) {
            m_disp.submitDeferredCommand<Audio::Settings::SetModelDefaultIn>(
                m, i);
          });

      in_count->setValue(m.getDefaultIn());
    }

    {
      out_count->setRange(0, 1024);
      lay->addRow(QObject::tr("Outputs"), out_count);
      QObject::connect(
          out_count,
          SignalUtils::QSpinBox_valueChanged_int(),
          w,
          [=, &m, &m_disp](int i) {
            m_disp.submitDeferredCommand<Audio::Settings::SetModelDefaultOut>(
                m, i);
          });

      out_count->setValue(m.getDefaultIn());
    }

    con(m, &Model::changed, w, [=, &m] {
      {
        auto val = m.getDefaultIn();
        if (val != in_count->value())
          in_count->setValue(val);
      }
      {
        auto val = m.getDefaultOut();
        if (val != out_count->value())
          out_count->setValue(val);
      }
    });
    return w;
  }
};
#endif
}
